import torch, math
from torch import nn


class AbstractPositionalEncoding(nn.Module):
    pass


class IdentityEncoding(AbstractPositionalEncoding):
    def __init__(self, *args, **kwargs):
        super().__init__()

    def forward(self, x):
        """
        Arguments:
            x: Tensor, shape ``[batch_size, seq_len, embedding_dim]``
        """
        return x


class ClassicPositionalEncoding(AbstractPositionalEncoding):

    def __init__(self, d_model: int, dropout: float = 0.1, max_len: int = 5000):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)

        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2)*(-math.log(10000.0)/d_model))
        pe = torch.zeros(1, max_len, d_model)
        pe[0, :, 0::2] = torch.sin(position*div_term)
        pe[0, :, 1::2] = torch.cos(position*div_term)

        self.register_buffer('pe', pe)

    def forward(self, x):
        """
        Arguments:
            x: Tensor, shape ``[batch_size, seq_len, embedding_dim]``
        """
        x = x + self.pe[:, :x.size(1), :]
        return self.dropout(x)


class PositionalAppender(AbstractPositionalEncoding):

    def __init__(self, d_model: int, dropout: float = 0.1, max_len: int = 5000):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)

        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2)*(-math.log(10000.0)/d_model))
        pe = torch.zeros(1, max_len, d_model)
        pe[0, :, 0::2] = torch.sin(position*div_term)
        pe[0, :, 1::2] = torch.cos(position*div_term)

        self.register_buffer('pe', pe)
        self.linear = nn.Linear(2*d_model, d_model)

    def forward(self, x):
        """
        Arguments:
            x: Tensor, shape ``[batch_size, seq_len, embedding_dim]``
        """

        appended = torch.cat((x, self.pe[:, :x.size(1), :].broadcast_to(x.shape)), dim=-1)
        return self.dropout(self.linear(appended))



import torch
from torch import nn

from BERTeam.networks.positional_encoder import IdentityEncoding, ClassicPositionalEncoding, PositionalAppender


class InputEmbedder(nn.Module):
    def forward(self, pre_embedding, preembed_mask):
        """
        Args:
            pre_embedding: (N, S, *) observation preembedding
            mask: (N, S) boolean mask, or None
        Returns:
            (embedding (N, S', E), mask (N, S'))
        """
        raise NotImplementedError


class DiscreteInputEmbedder(InputEmbedder):
    def __init__(self, num_embeddings, embedding_dim):
        super().__init__()
        self.embed = nn.Embedding(
            num_embeddings=num_embeddings,
            embedding_dim=embedding_dim,
        )
        self.pos_enc = IdentityEncoding()

    def forward(self, pre_embedding, preembed_mask):
        if pre_embedding is None:
            return None, None
        return self.pos_enc(self.embed(pre_embedding)), preembed_mask


class DiscreteInputPosEmbedder(DiscreteInputEmbedder):
    def __init__(self, num_embeddings, embedding_dim, dropout=.1):
        super().__init__(num_embeddings=num_embeddings, embedding_dim=embedding_dim)
        self.embed = nn.Embedding(
            num_embeddings=num_embeddings,
            embedding_dim=embedding_dim,
        )
        self.pos_enc = ClassicPositionalEncoding(d_model=embedding_dim, dropout=dropout)


class DiscreteInputPosAppender(DiscreteInputEmbedder):
    """
    instead of adding positional encoding, just appends it, and also has a linear layer to match dimensions
    """

    def __init__(self, num_embeddings, embedding_dim, dropout=.1):
        super().__init__(num_embeddings=num_embeddings, embedding_dim=embedding_dim)
        self.embed = nn.Embedding(
            num_embeddings=num_embeddings,
            embedding_dim=embedding_dim,
        )
        self.pos_enc = PositionalAppender(d_model=embedding_dim, dropout=dropout)


class LSTEmbedding(InputEmbedder):
    def __init__(self,
                 input_dim,
                 embedding_dim=512,
                 layers=4,
                 dropout=0,
                 device=None,
                 *args, **kwargs):
        super().__init__(*args, **kwargs)
        # skip batch and sequence dims
        self.flat = nn.Flatten(start_dim=2, end_dim=-1)
        self.lstm = nn.LSTM(input_size=input_dim,
                            hidden_size=embedding_dim,
                            num_layers=layers,
                            batch_first=True,
                            dropout=dropout,
                            bidirectional=False,
                            device=device,
                            )
        self.device = device
        self.embedding_dim = embedding_dim

    def subseqs_from_mask(self, mask):
        """
        splits a sequence according to where the mask is true
        Args:
            seq: (S,*) sequence
            mask: (S,) boolean mask of truth values
        Returns:
            iterable of (i,j) start end indices of each contiguous subsequence
        """
        out = []
        i = 0
        j = 0
        for j, value in enumerate(mask):
            if value:
                # this index is masked, so add the previous contiguous subsequence, if nonempty
                if i != j:
                    yield (i, j)

                # at best, next contiguous subsequence can start at next index
                i = j + 1
        # j is now len(mask)-1
        if i != j + 1:
            # if we end with an unmasked value, add the suffix in
            yield (i, j + 1)
        return out

    def output_from_lstm(self, seq, subseq=None):
        """
        gets output from a sequence
        Args:
            seq: (S,input_dim) sequence
            subseq: i,j start end
                uses 0, len(seq) by default
        Returns: (embed_dim,)
        """
        if subseq is not None:
            i, j = subseq
            seq = seq[i:j]
        out, _ = self.lstm.forward(seq)

        return out[-1]

    def forward(self, pre_embedding, preembed_mask):
        if pre_embedding is None:
            return None, None
        split = True
        if preembed_mask is None:
            split = False
            preembed_mask = torch.zeros(pre_embedding.shape[:2], dtype=torch.bool, device=self.device)

        batch_of_seq = self.flat(pre_embedding)
        outputs = []
        Sp = 0
        for seq, mask in zip(batch_of_seq, preembed_mask):
            # seq is (S,input_dim)
            # mask is (S,)
            if not split:
                outputs.append(self.output_from_lstm(seq=seq, subseq=None).reshape((1, -1)))
                Sp = max(Sp, 1)
            else:
                applied_lstm = [
                    self.output_from_lstm(seq=seq, subseq=subseq)
                    for subseq in self.subseqs_from_mask(mask=mask)
                ]

                outputs.append(torch.stack(applied_lstm, dim=0))
                Sp = max(Sp, len(applied_lstm))
        ret = torch.zeros((len(batch_of_seq), Sp, self.embedding_dim),
                          device=self.device)
        mask = torch.ones((len(batch_of_seq), Sp),
                          dtype=torch.bool,
                          device=self.device,
                          )
        for i, output in enumerate(outputs):
            ret[i, :len(output)] = output
            mask[i, :len(output)] = False
        return ret, mask



import torch, os
from torch import nn
from torch.utils.data import DataLoader

from BERTeam.buffer import LangReplayBuffer

from BERTeam.networks import TeamBuilder, BERTeam
from BERTeam.networks import DiscreteInputEmbedder, DiscreteInputPosEmbedder, DiscreteInputPosAppender
from BERTeam.networks import IdentityEncoding, ClassicPositionalEncoding, PositionalAppender


class TeamTrainer:
    def __init__(self, num_agents, MASK=-1):
        self.num_agents = num_agents
        self.MASK = MASK

    def reset_storage_dir(self, storage_dir):
        pass

    def clear(self):
        """
        clears any memory/data
        """
        pass

    def save(self, save_dir):
        pass

    def load(self, save_dir):
        pass

    def add_to_buffer(self, scalar, obs_preembed, team, obs_mask):
        """
        adds element to language replay buffer
        Args:
            scalar:
            obs_preembed:
            team:
            obs_mask:

        Returns:

        """
        pass

    def uniform_random_team(self, shape):
        return torch.randint(0, self.num_agents, shape)

    def get_total_distribution(self,
                               T,
                               N=1,
                               init_team=None,
                               tracked=None,
                               **kwargs,
                               ):
        raise NotImplementedError

    def get_member_distribution(self,
                                init_team,
                                indices=None,
                                **kwargs
                                ):
        """
        returns probability distribution of next member being added to team
        Args:
            init_team: already initialized team
            indices: indices to check, if None, checks all masked elements
        Returns:
            indices, distribution of agents for each index
                for this class, uniform
        """

        if indices is None:
            indices = torch.where(torch.eq(init_team, self.MASK))

        return indices, torch.ones((len(indices[0]), self.num_agents))/self.num_agents

    def add_member_to_team(self,
                           member,
                           T=None,
                           N=1,
                           init_team=None,
                           team_noise_model=None,
                           valid_locations=None,
                           **kwargs,
                           ):
        """
        adds the specified member to the team in a random position (sampled from the network distribution)
        one of T or init_team must be specified
        Args:
            member: member to add (index)
            T: team size
            N: number of teams to make (default 1)
            init_team: team to add member team
                if None, starts with all masks
            team_noise_model: noise model that takes T-element multinomial distributions and returns another (with noise)
                ((N,T) -> (N,T))
                default None
            valid_locations: boolean array of size (N,T) that determines whether each location is valid
        Returns:
            team with member attached
        """
        if init_team is None:
            init_team = self.create_masked_teams(T=T, N=N)

        # (N, T, num_agents)
        dist = torch.ones((N, T, self.num_agents))/self.num_agents

        # (N, T)
        conditional_dist = dist[:, :, member]

        if team_noise_model is not None:
            conditional_dist = team_noise_model(conditional_dist)

        # set all non-masked or invalid entries to 0
        conditional_dist[torch.where(torch.not_equal(init_team, self.MASK))] = 0
        if valid_locations is not None:
            conditional_dist = conditional_dist*valid_locations  # sets invalid locations to 0

        valid_indices = torch.where(torch.sum(conditional_dist, axis=1) > 0)[0]
        if len(valid_indices) == 0:
            # there are no mask tokens, or the conditional distribution is all zeros
            return init_team

        # samples from conditional_dist at the valid indices
        places = torch.multinomial(conditional_dist[valid_indices,], 1).flatten()

        init_team[valid_indices, places] = member
        return init_team

    def fill_in_teams(self,
                      initial_teams,
                      noise_model=None,
                      num_masks=None,
                      valid_members=None,
                      **kwargs
                      ):
        """
        replaces all [MASK] with team members
            repeatedly calls mutate_add_member
        Args:
            initial_teams: initial torch array of team members, shape (N,T)
            obs_preembed: input to give input embedder, size (N,S,*)
                None if no input
            obs_mask: size (N,S) boolean array of whether to mask each input embedding
            noise_model: noise to add to probability distribution, if None, doesnt add noise
            num_masks: specify the largest number of [MASK] tokens in any row
                if None, just calls mutate_add_member (T) times
            valid_members: (N,T,num_agents) boolean array of which agents are valid for which locations
        Returns:
            filled in teams of size (N,T)
        """
        N, T = initial_teams.shape
        if num_masks is None:
            num_masks = T
        for _ in range(num_masks):
            initial_teams = self.mutate_add_member(initial_teams=initial_teams,
                                                   indices=None,
                                                   noise_model=noise_model,
                                                   valid_members=valid_members,
                                                   **kwargs,
                                                   )
        return initial_teams

    def mutate_add_member(self,
                          initial_teams,
                          indices=None,
                          noise_model=None,
                          valid_members=None,
                          **kwargs,
                          ):
        """
        updates initial teams by updating specified indices with samples from the probability distribution
            if indices is None, chooses indices that correspond with [MASK] tokens (one for each row)
        Args:
            initial_teams: initial torch array of team members, shape (N,T)
            indices: indices to replace (if None, picks one masked index at random)
                should be in pytorch format, (list of dim 0 indices, list of dim 1 indices)
            obs_preembed: input to give input embedder, size (N,S,*)
                None if no input
            obs_mask: size (N,S) boolean array of whether to mask each input embedding
            noise_model: noise to add to probability distribution, if None, doesnt add noise
            valid_members: (N,T,num_agents) boolean array of which agents are valid for which locations
        Returns:
            team with updates
        """
        N, T = initial_teams.shape
        if indices is None:
            indices = [[], []]
            for i in range(N):
                potential = torch.where(torch.eq(initial_teams[i], self.MASK))[0]
                if len(potential) > 0:
                    indices[0].append(i)
                    indices[1].append(potential[torch.randint(0, len(potential), (1,))])
                # else:
                #    print("warning: tried to add to full team")
        if len(indices[0]) == 0:
            # no [MASK] tokens exist
            return initial_teams
        dist = torch.ones((N, T, self.num_agents))/self.num_agents

        if noise_model is not None:
            # add noise if this is a thing
            dist = noise_model(dist)

        if valid_members is not None:
            dist = dist*valid_members
        # just look at the relevant indices
        dist = dist[indices]

        # torch.multinomial samples each

        initial_teams[indices] = torch.multinomial(dist, 1).flatten()
        return initial_teams

    def create_masked_teams(self, T, N=1):
        """
        Args:
            T: team size
            N: batch size
        Returns:
            an (N,T) tensor of all masked members
        """
        return torch.fill(torch.zeros((N, T), dtype=torch.long), self.MASK)

    def train(self, *args, **kwargs):
        pass

    def create_nose_model_towards_uniform(self, t):
        """
        creates a noise model that takes in a distribution and does a weighted average with a uniform dist
        Args:
            t: weight to give uniform dist (1.0 replaces the distribution, 0.0 leaves dist unchanged)
                or array of length (K,), doing it on individual values
        Returns:
            distribution -> distribution map, which are sized (N,K) for an N batch of a K-multinomial
        """
        if isinstance(t, (int, float)):
            return (lambda dist: (1 - t)*dist + (t)*1./dist.shape[1])
        else:
            return (lambda dist:
                    ((1 - t)*dist + (t)*torch.ones_like(dist)/dist.shape[1])/
                    (torch.sum((1 - t)*dist) + torch.sum(t)*1/dist.shape[1])
                    )
            # equivalent to
            # return (lambda dist:
            #        ((1 - t)*dist + (t)*torch.ones_like(dist)/dist.shape[1])/
            #        torch.sum((1 - t)*dist + (t)*torch.ones_like(dist)/dist.shape[1])
            #        )

    def create_teams(self,
                     T,
                     N=1,
                     obs_preembed=None,
                     obs_mask=None,
                     noise_model=None,
                     ):

        """
        creates random teams of size (N,T)
        Args:
            T: number of team members
            N: number of teams to make (default 1)
            obs_preembed: input to give input embedder, size (N,S,*)
                None if no input
            obs_mask: size (N,S) boolean array of whether to mask each input embedding
            noise_model: noise to add to probability distribution, if None, doesnt add noise
        Returns:
            filled in teams of size (N,T)
        """
        return self.uniform_random_team(shape=(N, T))


class MLMTeamTrainer(TeamTrainer):
    def __init__(self,
                 team_builder: TeamBuilder,
                 buffer: LangReplayBuffer,
                 storage_dir=None,
                 ):
        """
        Args:
            team_builder:
            buffer:
            storage_dir: place to store files
                if None, sets later
        """
        super().__init__(
            num_agents=team_builder.berteam.num_agents,
            MASK=team_builder.berteam.MASK,
        )
        self.team_builder = team_builder
        self.buffer = buffer

        if storage_dir is not None:
            self.reset_storage_dir(storage_dir=storage_dir)

        self.optim = torch.optim.Adam(team_builder.parameters())

    def reset_storage_dir(self, storage_dir):
        self.buffer.reset_storage_dir(storage_dir=os.path.join(storage_dir, 'buffer'))

    def clear(self):
        """
        clears any memory/data
        """
        super().clear()
        self.buffer.clear()

    def save(self, save_dir):
        super().save(save_dir=save_dir)
        self.buffer.save(os.path.join(save_dir, 'buffer'))
        dic = {'model': self.team_builder.state_dict(),
               'optim': self.optim.state_dict(),
               }
        torch.save(dic, os.path.join(save_dir, 'model.pkl'))

    def load(self, save_dir):
        super().load(save_dir=save_dir)
        self.buffer.load(os.path.join(save_dir, 'buffer'))
        dic = torch.load(os.path.join(save_dir, 'model.pkl'))
        self.team_builder.load_state_dict(dic['model'])
        self.optim.load_state_dict(dic['optim'])

    def add_to_buffer(self, scalar, obs_preembed, team, obs_mask):
        item = (scalar, obs_preembed, team, obs_mask)
        self.buffer.push(item=item)

    def train(self,
              batch_size,
              minibatch=None,
              mask_probs=None,
              replacement_probs=(.8, .1, .1),
              mask_obs_prob=.1,
              ):
        if minibatch is None:
            minibatch = batch_size
        if not len(self.buffer):
            print('WARNING: trying to train on empty buffer')
            return None
        loss = 0
        for i in range(0, batch_size, minibatch):
            tinybatch = min(i + minibatch, batch_size) - i
            # cut off the last batch if we go over
            step_loss = self.training_step(
                batch_size=tinybatch,
                mask_probs=mask_probs,
                replacement_probs=replacement_probs,
                mask_obs_prob=mask_obs_prob,
            )
            loss += tinybatch*step_loss
        # clear any remaining gradient from memory
        self.optim.zero_grad()

        # torch.clear_autocast_cache()
        # if torch.cuda.is_available():
        #    torch.cuda.empty_cache()
        return loss/batch_size

    def training_step(self,
                      batch_size,
                      mask_probs=None,
                      replacement_probs=(.8, .1, .1),
                      mask_obs_prob=.1,
                      ):
        """
        Args:
            batch_size: size of batch to train on
            mask_probs: list of proabilities of masking to sample from
                if None, samples (1/team_size, 2/team_size, ..., 1)
                    for each obtained team size
            replacement_probs: proportion of ([MASK], random element, same element) to replace masked elements with
                default (.8, .1, .1) because BERT
            scalar: thing to multiply losses
                should be 1 for normal MLM training
                    -1 to push model away from 'bad' teamas
            mask_obs_prob: if >0, randomly masks observations with this probability as well
        Returns:
            avg losses for whole dataset
        """
        if not len(self.buffer):
            print('WARNING: trying to train on empty buffer')
            return None
        data = self.buffer.sample(batch=batch_size)
        self.optim.zero_grad()
        losses = torch.zeros(1)
        count = 0
        for (scalar, obs_preembed, team, obs_mask) in data:
            team = team.view((1, -1))
            if torch.is_tensor(obs_preembed) and torch.all(torch.isnan(obs_preembed)):
                obs_preembed = None
            if torch.is_tensor(obs_mask) and torch.all(torch.isnan(obs_mask)):
                obs_mask = None
            if mask_probs is None:
                N, T = team.shape
                mask_probs = torch.arange(0, T + 1)/T

            mask_prob = mask_probs[torch.randint(0, len(mask_probs), (1,))]

            loss = self._mask_and_learn(obs_preembed=obs_preembed,
                                        teams=team,
                                        obs_mask=obs_mask,
                                        mask_prob=mask_prob,
                                        replacement_probs=replacement_probs,
                                        mask_obs_prob=mask_obs_prob,
                                        )
            # keep gradients
            losses += loss
            count += 1
        mean_loss = losses/count
        mean_loss.backward()
        self.optim.step()
        return mean_loss.item()

    def _mask_and_learn(self,
                        obs_preembed,
                        teams,
                        obs_mask,
                        mask_prob,
                        replacement_probs=(.8, .1, .1),
                        mask_obs_prob=.1,
                        ):
        """
        runs learn step on a lot of mask_probabilities
            need to test with a wide range of mask probabilites because for generation, we may start with a lot of masks
        Args:
            obs_preembed: tensor (N, S, *) of input preembeddings, or None if no preembeddings
            teams: tensor (N, T) of teams
            obs_mask: boolean tensor (N, S) of whether to pad each input
            mask_probs: list of proabilities of masking to use
                if None, uses (1/team_size, 2/team_size, ..., 1)
            replacement_probs: proportion of ([MASK], random element, same element) to replace masked elements with
                default (.8, .1, .1) because BERT
            mask_obs_prob: if >0, randomly mask observations with this prob
        Returns:
            avg crossentropy loss
        """
        if mask_obs_prob > 0 and obs_preembed is not None:
            if obs_mask is not None:
                temp_obs_mask = obs_mask.clone()
            else:
                temp_obs_mask = torch.zeros((obs_preembed.shape[:2]))
            additional_mask = torch.rand_like(temp_obs_mask, dtype=torch.float) < mask_obs_prob
            temp_obs_mask = torch.logical_or(temp_obs_mask, additional_mask)
        else:
            temp_obs_mask = obs_mask
        loss = self._get_losses(obs_preembed=obs_preembed,
                                teams=teams,
                                obs_mask=temp_obs_mask,
                                mask_prob=mask_prob,
                                replacement_probs=replacement_probs,
                                )
        return loss

    def _get_losses(self,
                    obs_preembed,
                    teams,
                    obs_mask,
                    mask_prob=.5,
                    replacement_probs=(.8, .1, .1),
                    ):
        """
        randomly masks winning team members, runs the transformer token prediction model, and gets crossentropy loss
            of predicting the masked tokens
        Args:
            obs_preembed: tensor (N, S, *) of input preembeddings, or None if no preembeddings
            teams: tensor (N, T) of teams
            obs_mask: boolean tensor (N, S) of whether to pad each input
            mask_prob: proportion of elements to mask (note that we will always mask at least one per batch
            replacement_probs: proportion of ([MASK], random element, same element) to replace masked elements with
                default (.8, .1, .1) because BERT
        Returns:
            crossentropy loss of prediction
        """

        in_teams, mask, mask_indices = self._randomly_mask_teams(teams=teams,
                                                                 mask_prob=mask_prob,
                                                                 replacement_props=replacement_probs,
                                                                 )
        logits = self.team_builder.forward(obs_preembed=obs_preembed,
                                           target_team=in_teams,
                                           obs_mask=obs_mask,
                                           output_probs=True,
                                           pre_softmax=True,
                                           )
        criterion = nn.CrossEntropyLoss()

        loss = criterion(logits[mask_indices], teams[mask_indices])
        return loss

    def _randomly_mask_teams(self, teams, mask_prob, replacement_props):
        """
        Args:
            teams: size (N,T) tensor to randomly mask
            mask_prob: proportion of elements to mask (note that we will always mask at least one per batch)
            replacement_props: proportion of ([MASK], random element, same element) to replace masked elements with
                tuple of three floats
        """
        masked_teams = teams.clone()
        N, T = teams.shape

        which_to_mask = torch.bernoulli(torch.ones_like(teams)*mask_prob)
        forced_mask = torch.randint(0, T, (N,))
        which_to_mask[torch.arange(N), forced_mask] = 1

        mask_indices = torch.where(torch.eq(which_to_mask, 1))
        num_masks = len(mask_indices[0])
        what_to_replace_with = torch.multinomial(torch.tensor([replacement_props for _ in range(num_masks)]),
                                                 1).flatten()
        # this will be a tensor of 0s, 1s, and 2s, indicating whether to replace each mask with
        #   [MASK], a random element, or not to replace it

        # map this to -1,-2,-3 respectively
        what_to_replace_with = -1 - what_to_replace_with
        # now map to the actual values to replace with in the array
        what_to_replace_with = torch.masked_fill(what_to_replace_with,
                                                 mask=torch.eq(what_to_replace_with, -1),
                                                 value=self.MASK,
                                                 )
        rand_rep = torch.where(torch.eq(what_to_replace_with, -2))
        what_to_replace_with[rand_rep] = self.uniform_random_team((len(rand_rep[0]),))

        unchange = torch.where(torch.eq(what_to_replace_with, -3))
        # grab the correct team members from the original array
        what_to_replace_with[unchange] = masked_teams[[dim_idx[unchange] for dim_idx in mask_indices]]

        masked_teams[mask_indices] = what_to_replace_with
        return masked_teams, which_to_mask, mask_indices

    def add_member_to_team(self,
                           member,
                           T=None,
                           N=1,
                           init_team=None,
                           obs_preembed=None,
                           obs_mask=None,
                           team_noise_model=None,
                           valid_locations=None,
                           ):
        """
        adds the specified member to the team in a random position (sampled from the network distribution)
        one of T or init_team must be specified
        Args:
            member: member to add (index)
            T: team size
            N: number of teams to make (default 1)
            init_team: team to add member team
                if None, starts with all masks
            obs_preembed: observation to condition on
            obs_mask: mask for observations
            team_noise_model: noise model that takes T-element multinomial distributions and returns another (with noise)
                ((N,T) -> (N,T))
                default None
            valid_locations: boolean array of size (N,T) that determines whether each location is valid
        Returns:
            team with member attached
        """
        if init_team is None:
            init_team = self.create_masked_teams(T=T, N=N)

        # (N, T, num_agents)
        dist = self.team_builder.forward(obs_preembed=obs_preembed,
                                         target_team=init_team,
                                         obs_mask=obs_mask,
                                         output_probs=True,
                                         pre_softmax=False,
                                         )
        # (N, T)
        conditional_dist = dist[:, :, member]

        if team_noise_model is not None:
            conditional_dist = team_noise_model(conditional_dist)

        # set all non-masked entries to 0
        conditional_dist[torch.where(init_team != self.MASK)] = 0
        if valid_locations is not None:
            conditional_dist = conditional_dist*valid_locations

        valid_indices = torch.where(torch.sum(conditional_dist, axis=1) > 0)[0]
        if len(valid_indices) == 0:
            # there are no mask tokens, or the conditional distribution is all zeros
            return init_team

        # samples from conditional_dist at the valid indices
        places = torch.multinomial(conditional_dist[valid_indices,], 1).flatten()

        init_team[valid_indices, places] = member
        return init_team

    def create_teams(self,
                     T,
                     N=1,
                     obs_preembed=None,
                     obs_mask=None,
                     noise_model=None,
                     ):

        """
        creates random teams of size (N,T)
        Args:
            T: number of team members
            N: number of teams to make (default 1)
            obs_preembed: input to give input embedder, size (N,S,*)
                None if no input
            obs_mask: size (N,S) boolean array of whether to mask each input embedding
            noise_model: noise to add to probability distribution, if None, doesnt add noise
        Returns:
            filled in teams of size (N,T)
        """
        return self.fill_in_teams(initial_teams=self.create_masked_teams(T=T, N=N),
                                  obs_preembed=obs_preembed,
                                  obs_mask=obs_mask,
                                  noise_model=noise_model,
                                  )

    def get_total_distribution(self,
                               T,
                               N=1,
                               init_team=None,
                               prob=1,
                               tracked=None,
                               obs_preembed=None,
                               obs_mask=None,
                               noise_model=None,
                               valid_members=None,
                               ):
        if init_team is None:
            init_team = self.create_masked_teams(T=T, N=N)
        if tracked is None:
            tracked = dict()

        indices, dist = self.get_member_distribution(init_team=init_team,
                                                     indices=None,
                                                     obs_preembed=obs_preembed,
                                                     obs_mask=obs_mask,
                                                     noise_model=noise_model,
                                                     valid_members=valid_members,
                                                     )
        if indices is None:
            obj = tuple(init_team.detach().numpy().flatten())
            if obj not in tracked:
                tracked[obj] = 0
            tracked[obj] += prob.item()
        else:
            for i, idx in enumerate(zip(*indices)):
                for member, value in enumerate(dist[i]):
                    temp_team = init_team.clone()
                    temp_team[idx] = member
                    # probability of sleceting this is prob of selecting init_team
                    # times prob of selecting this index (1/ number of indices)
                    # times transformer prob

                    selection_prob = value*prob/len(indices[0])
                    self.get_total_distribution(T=T,
                                                N=N,
                                                init_team=temp_team,
                                                prob=selection_prob,
                                                tracked=tracked,
                                                obs_preembed=obs_preembed,
                                                obs_mask=obs_mask,
                                                noise_model=noise_model,
                                                valid_members=valid_members,
                                                )
        return tracked

    def get_member_distribution(self,
                                init_team,
                                indices=None,
                                obs_preembed=None,
                                obs_mask=None,
                                noise_model=None,
                                valid_members=None,
                                ):
        if indices is None:
            indices = torch.where(torch.eq(init_team, self.MASK))
        if len(indices[0]) == 0:
            return None, None
        output = self.team_builder.forward(obs_preembed=obs_preembed,
                                           target_team=init_team,
                                           obs_mask=obs_mask,
                                           output_probs=True,
                                           pre_softmax=False,
                                           )

        dist = output[indices]  # (|indices|, num_agents) multinomial distribution for each index to update
        if noise_model is not None:
            # add noise if this is a thing
            dist = noise_model(dist)

        if valid_members is not None:
            # set invalid members to 0
            dist = dist*(valid_members[indices])
        return indices, dist

    def mutate_add_member(self,
                          initial_teams,
                          indices=None,
                          obs_preembed=None,
                          obs_mask=None,
                          noise_model=None,
                          valid_members=None,
                          ):
        """
        updates initial teams by updating specified indices with samples from the probability distribution
            if indices is None, chooses indices that correspond with [MASK] tokens (one for each row)
        Args:
            initial_teams: initial torch array of team members, shape (N,T)
            indices: indices to replace (if None, picks one masked index at random)
                should be in pytorch format, (list of dim 0 indices, list of dim 1 indices)
            obs_preembed: input to give input embedder, size (N,S,*)
                None if no input
            obs_mask: size (N,S) boolean array of whether to mask each input embedding
            noise_model: noise to add to probability distribution, if None, doesnt add noise
            valid_members: (N,T,num_agents) boolean array of which agents are valid for which locations
        Returns:
            team with updates
        """
        N, T = initial_teams.shape
        if indices is None:
            indices = [[], []]
            for i in range(N):
                potential = torch.where(initial_teams[i] == self.MASK)[0]
                if len(potential) > 0:
                    indices[0].append(i)
                    indices[1].append(potential[torch.randint(0, len(potential), (1,))])
                # else:
                #    print("warning: tried to add to full team")
        if len(indices[0]) == 0:
            # no [MASK] tokens exist
            return initial_teams
        indices, dist = self.get_member_distribution(init_team=initial_teams,
                                                     indices=indices,
                                                     obs_preembed=obs_preembed,
                                                     obs_mask=obs_mask,
                                                     noise_model=noise_model,
                                                     valid_members=valid_members
                                                     )

        # torch.multinomial samples each
        initial_teams[indices] = torch.multinomial(dist, 1).flatten()
        return initial_teams


class DiscreteInputTrainer(MLMTeamTrainer):
    def __init__(self,
                 buffer: LangReplayBuffer,
                 num_agents,
                 num_input_tokens,
                 embedding_dim=512,
                 nhead=8,
                 num_encoder_layers=16,
                 num_decoder_layers=16,
                 dim_feedforward=None,
                 dropout=.1,
                 pos_encode_input=True,
                 append_pos_encode_input=False,
                 pos_encode_teams=True,
                 append_pos_encode_teams=False,
                 ):
        if pos_encode_teams:
            if append_pos_encode_teams:
                PosEncConstructorTeams = PositionalAppender
            else:
                PosEncConstructorTeams = ClassicPositionalEncoding
        else:
            PosEncConstructorTeams = IdentityEncoding
        berteam = BERTeam(num_agents=num_agents,
                          embedding_dim=embedding_dim,
                          nhead=nhead,
                          num_encoder_layers=num_encoder_layers,
                          num_decoder_layers=num_decoder_layers,
                          dim_feedforward=dim_feedforward,
                          dropout=dropout,
                          PosEncConstructor=PosEncConstructorTeams,
                          )
        if pos_encode_input:
            if append_pos_encode_input:
                Constructor = DiscreteInputPosAppender
            else:
                Constructor = DiscreteInputPosEmbedder
            input_embedder = Constructor(num_embeddings=num_input_tokens,
                                         embedding_dim=embedding_dim,
                                         dropout=dropout,
                                         )
        else:
            input_embedder = DiscreteInputEmbedder(num_embeddings=num_input_tokens,
                                                   embedding_dim=embedding_dim,
                                                   )

        super().__init__(
            buffer=buffer,
            team_builder=TeamBuilder(
                berteam=berteam,
                input_embedder=input_embedder,
            )
        )


if __name__ == '__main__':
    from matplotlib import pyplot as plt

    N = 32  # number of teams to train on per epoch
    eval_N = 9  # number of teams to evaluate on
    S = 2  # sequence length of input
    T = 5  # size of team
    num_inputs = 4  # number of distinct input tokens

    epochs = 300
    torch.random.manual_seed(69)

    E = 64  # embedding dim

    test = DiscreteInputTrainer(num_agents=69,
                                num_input_tokens=num_inputs,
                                embedding_dim=E,
                                pos_encode_input=True,
                                append_pos_encode_input=True,
                                pos_encode_teams=True,
                                append_pos_encode_teams=True,
                                num_decoder_layers=3,
                                num_encoder_layers=3,
                                dropout=0.1,
                                nhead=8,
                                )

    basic_team = torch.arange(T, dtype=torch.long)


    def random_shuffle():
        return basic_team[torch.randperm(len(basic_team))]


    def correct_output(input_pre):
        # test it on shifted inputs, mod some number
        # return (basic_team + input_pre[0])%4
        return (basic_team + torch.sum(input_pre))%4
        # return basic_team


    losses = []
    for epoch in range(epochs):
        # input_preembedding_gen = (torch.randint(0, num_inputs, (i%S,)) if i > 0 else None for i in range(N))
        # input_embedding_gen = (None for i in range(N))

        input_mask = [torch.nan for _ in range(N)]
        input_preembedding = torch.randint(0, num_inputs, (N, S))
        # input_mask = torch.zeros((N, S), dtype=torch.bool)
        # out_teams = torch.stack([random_shuffle() for _ in range(N)], dim=0)
        out_teams = torch.stack([correct_output(t) for t in input_preembedding], dim=0)

        data = list(zip(input_preembedding, out_teams, input_mask))
        loader = DataLoader(data,
                            shuffle=True,
                            batch_size=64,
                            )
        loss = test.training_step(
            data=loader,
            sgd=True
        )

        losses.append(loss)
        print('epoch', epoch, '\tloss', loss.item())
    # print(test.mask_and_learn(input_embedding_gen=input_embeddings,
    #                          winning_team_gen=(init_team.clone() for _ in range(N))))
    init_teams = test.create_masked_teams(T=T, N=eval_N)
    print(init_teams)
    # input_pre = torch.arange(0, num_teams).view((-1, 1))%num_inputs

    input_preembedding = torch.randint(0, num_inputs, (eval_N, S))
    for i in range(T):
        # force it to add member i to the team
        # init_teams = test.add_member_to_team(i, init_team=init_teams, obs_preembed=input_preembedding)

        test.mutate_add_member(initial_teams=init_teams, obs_preembed=input_preembedding)
        print(init_teams)
    print('goal:')
    goal = torch.stack([correct_output(t) for t in input_preembedding])
    print(goal)

    print('accuracy:',
          round((torch.sum(init_teams == goal)/torch.sum(torch.ones_like(init_teams))).item()*100, 2),
          '%')
    plt.plot(range(len(losses)), losses)
    plt.show()



import torch, os, sys, shutil
import dill as pickle
from BERTeam.outcome import PlayerInfo


class LangReplayBuffer:
    storage_dir = None

    def reset_storage_dir(self, storage_dir):
        self.storage_dir = storage_dir
        if not os.path.exists(storage_dir):
            os.makedirs(storage_dir)
        self.reset_storage()

    def reset_storage(self):
        """
        resets internal buffer
        """
        raise NotImplementedError

    def extend(self, items):
        for item in items:
            self.push(item)

    def add_outcome(self,
                    teams,
                    outcome,
                    filter=None,
                    ignore_observation=False,
                    chain_observation=True):
        """
        adds result of Outcome.get_outcome to buffer
        Args:
            teams: list of teams
            outcome:
                list corresponding to teams
                    [
                        outcome score,
                        list corresponding to players of PlayerInfo(
                            obs_preembed=player observation (None or size (S,*) seq of observations);
                            obs_mask=observation mask (None or size (S,) boolean array of which items to mask;
                            )
                        list can be empty, this will correspond to an empty observation
                    ]
            filter: whether to add each outcome, bsed on score
        """

        for team, (score, playerinfos) in zip(teams, outcome):
            if filter is not None:
                if not filter(score):
                    continue
            playerinfo = PlayerInfo()

            if not ignore_observation and not chain_observation:
                # in this case, add one example for each player
                for pi in playerinfos:
                    obs_preembed, obs_mask = pi.get_data()
                    item = (score, obs_preembed, team, obs_mask)
                    self.push(item)
                continue
            # otherwise, we combine the observations (if not ignored)
            if chain_observation and not ignore_observation:
                for pi in playerinfos:
                    playerinfo = playerinfo.union_obs(other_player_info=pi)
            obs_preembed, obs_mask = playerinfo.get_data()
            item = (score, obs_preembed, team, obs_mask)
            self.push(item)

    def push(self, item):
        """
        pushes an item into replay buffer
        Args:
            item: item
        Returns: item that is displaced, or None if no such item
        """
        raise NotImplementedError

    def sample_one(self):
        raise NotImplementedError

    def clear(self):
        pass

    def save(self, save_dir):
        pass

    def load(self, save_dir):
        pass

    def sample(self, batch, **kwargs):
        for _ in range(batch):
            yield self.sample_one()

    # def __getitem__(self, item):
    #    raise NotImplementedError

    def __len__(self):
        raise NotImplementedError


class ReplayBufferDiskStorage(LangReplayBuffer):
    def __init__(self,
                 storage_dir=None,
                 capacity=1e6,
                 device=None,
                 ):
        self.idx = 0
        self.size = 0
        self.capacity = capacity
        self.device = device
        if storage_dir is not None:
            self.reset_storage_dir(storage_dir=storage_dir)

    def clear(self):
        super().clear()
        if self.storage_dir is not None:
            if os.path.exists(self.storage_dir):
                shutil.rmtree(self.storage_dir)

    def save(self, save_dir):
        super().save(save_dir=save_dir)
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)
        shutil.copytree(src=self.storage_dir, dst=save_dir)

    def load(self, save_dir):
        super().load(save_dir=save_dir)
        self.clear()
        shutil.copytree(src=save_dir, dst=self.storage_dir)
        self.load_place(force=False)

    def reset_storage(self):
        self.clear()
        os.makedirs(self.storage_dir)
        self.size = 0
        self.idx = 0
        self.save_place()

    def save_place(self):
        """
        saves idx and size to files as well
        """
        pickle.dump(
            {
                'size': self.size,
                'idx': self.idx,
            },
            open(self._get_file('info'), 'wb')
        )

    def load_place(self, force=False):
        info_file = self._get_file(name='info')
        if os.path.exists(info_file):
            dic = pickle.load(open(info_file, 'rb'))
            self.size = dic['size']
            self.idx = dic['idx']
        else:
            if force:
                print('failed to load file:', info_file)
                print('resetting storage')
                self.reset_storage()
            else:
                raise Exception('failed to load file: ' + info_file)

    def _get_file(self, name):
        return os.path.join(self.storage_dir, str(name) + '.pkl')

    def push(self, item):
        if self.size == self.capacity:
            disp = self.__getitem__(self.idx)
        else:
            disp = None
        pickle.dump(item, open(self._get_file(self.idx), 'wb'))

        self.size = max(self.idx + 1, self.size)
        self.idx = int((self.idx + 1)%self.capacity)

        self.save_place()
        return disp

    def _grab_item_by_idx(self, idx, change_device=True):
        item = pickle.load(open(self._get_file(name=idx), 'rb'))
        return self._convert_device(item=item, change_device=change_device)

    def _convert_device(self, item, change_device):
        if change_device:
            if type(item) == tuple:
                item = tuple(self._convert_device(t, change_device=change_device)
                             for t in item)
            elif torch.is_tensor(item):
                item = item.to(self.device)
        return item

    def sample_one(self):
        return self[torch.randint(0, self.size, (1,))]

    def __getitem__(self, item):
        if item >= self.size:
            raise IndexError
        return self._grab_item_by_idx(idx=int((self.idx + item)%self.size))

    def __len__(self):
        return self.size


class BinnedReplayBufferDiskStorage(LangReplayBuffer):
    """
    creates multiple disk replay buffers, each representing a 'bin' of data
    items must begin with a scalar that represents how much it should show up in the data
    """

    def __init__(self,
                 storage_dir=None,
                 bounds=None,
                 capacity=1e6,
                 device=None,
                 ):
        """
        Args:
            bounds: sorted list of numbers to divide the input into bins based on the scalar associated with it
                each bin keeps track of the average value
                upon sampling, an extra parameter can be added to determine with what
                    frequency elements from each bin appear

                values <=bounds[0] or >bounds[1] are ignored
                by default, uses [1/2, 1], corresponding to one bin of range (1/2,1]
                    This is intended for a 2 player game where we want to capture outcomes that are better than ties
                        Usually, in an n-player game, the bounds should range [1/n,...,1] to capture outcomes that are
                            better than ties
        """
        super().__init__()
        if bounds is None:
            bounds = [1/2, 1]
        self.bins = [ReplayBufferDiskStorage(storage_dir=None,
                                             capacity=capacity,
                                             device=device)
                     for i in range(len(bounds) - 1)
                     ]
        self.info = {
            'bounds': tuple(bounds),
            'avgs': torch.zeros(len(bounds) - 1),
            'size': 0,
        }
        if storage_dir is not None:
            self.reset_storage_dir(storage_dir=storage_dir)
        self.weights = None

    def reset_storage_dir(self, storage_dir):
        super().reset_storage_dir(storage_dir=storage_dir)
        for i, biin in enumerate(self.bins):
            biin.reset_storage_dir(storage_dir=os.path.join(storage_dir, 'bin_' + str(i)))

    def reset_storage(self):
        self.clear()
        os.makedirs(self.storage_dir)
        self.idx = 0

    def set_size(self, size):
        self.info['size'] = size

    @property
    def size(self):
        return self.info['size']

    @property
    def bounds(self):
        return self.info['bounds']

    @property
    def avgs(self):
        return self.info['avgs']

    def bin_search(self, value, possible=None):
        """
        binary bin search
        Args:
            value: value to put into bin
            possible: possible bindices (if None, searches all)
        Returns:
            i such that self.bounds[i] < value<= self.bounds[i+1]
            or None if not possible
        """
        if possible is None:
            possible = (0, len(self.bounds) - 1)
        i, j = possible
        if i + 1 == j:
            if self.bounds[i] < value and value <= self.bounds[i + 1]:
                return i
            else:
                return None
        mid = (i + j)//2
        # i+1 <= mid <= j-1
        if self.bounds[mid] < value:
            return self.bin_search(value, (mid, j))
        if value < self.bounds[mid]:
            return self.bin_search(value, (i, mid))
        if value == self.bounds[mid]:
            # 0 <= i <= mid-1
            return mid - 1

    def push(self, item):
        scalar = item[0]
        biin = self.bin_search(scalar)
        if biin is not None:
            disp = self.bins[biin].push(item)
            if disp is None:
                rem = 0
                self.set_size(self.size + 1)
            else:
                rem = disp[0]

            self.avgs[biin] = (self.avgs[biin]*(len(self.bins[biin]) - 1) + scalar - rem)/len(self.bins[biin])

    def set_weights(self, values_to_weights):
        """
        sets weights of each  bin according to the average value of its elements
        Args:
        values_to_weights: a function (tensor -> tensor) ([0,1] -> R+), weights to give a bin with a particular value
            if None, uses weights of len(bin) for each bin (this corresponds to uniformly sampling an element)
        """
        self.weights = values_to_weights(self.avgs)

    def sample_one(self):
        """
        samples one according to weights set
        if weights are not set, samples uniformly from seen examples
        first samples a bin according to weights, then samples an element from bin
        """
        weights = self.weights
        if weights is None:
            weights = self.bin_lens
        # weight empty bins at 0
        weights = weights*(self.bin_lens > 0)

        biin = torch.multinomial(weights, 1)
        return self.bins[biin].sample_one()

    def sample(self, batch, **kwargs):
        """
        samples a bin according to the average value of its elements, then samples elements from the bin
        Args:
            batch: number of elements to sample
            **kwargs:
        Returns:
            Iterable of samples
        """
        for item in super().sample(batch=batch, **kwargs):
            yield item

    def clear(self):
        super().clear()
        for guy in self.bins:
            guy.clear()
        if self.storage_dir is not None:
            if os.path.exists(self.storage_dir):
                shutil.rmtree(self.storage_dir)

    def save(self, save_dir):
        super().save(save_dir=save_dir)
        if os.path.exists(save_dir):
            shutil.rmtree(save_dir)
        os.makedirs(save_dir)
        f = open(os.path.join(save_dir, 'info.pkl'), 'wb')
        pickle.dump(self.info, f)
        f.close()

        for i, guy in enumerate(self.bins):
            guy.save(save_dir=os.path.join(save_dir, 'bin_' + str(i)))

    def load(self, save_dir):
        super().load(save_dir=save_dir)
        f = open(os.path.join(save_dir, 'info.pkl'), 'rb')
        self.info.update(pickle.load(f))
        f.close()

        for i, guy in enumerate(self.bins):
            guy.load(save_dir=os.path.join(save_dir, 'bin_' + str(i)))

    def __len__(self):
        return self.size

    def __getitem__(self, item):
        if item >= self.size:
            raise IndexError
        for biin in self.bins:
            if item < biin.size:
                return biin[item]
            item -= biin.size

    @property
    def bin_lens(self):
        return torch.tensor([len(bin) for bin in self.bins],
                            dtype=torch.float,
                            )


if __name__ == '__main__':
    DIR = os.path.dirname(os.path.dirname(os.path.join(os.getcwd(), sys.argv[0])))
    test = ReplayBufferDiskStorage(capacity=3, storage_dir=os.path.join(DIR, 'data', 'buffers', 'replay_buffer_test'))
    test.extend('help')
    print(list(test.sample(3)))



import torch
from torch import nn
from torch.nn import Transformer, Embedding

from BERTeam.networks.positional_encoder import IdentityEncoding, ClassicPositionalEncoding, PositionalAppender
from BERTeam.networks.input_embedding import InputEmbedder


class BERTeam(nn.Module):
    def __init__(self,
                 num_agents,
                 embedding_dim=512,
                 nhead=8,
                 num_encoder_layers=16,
                 num_decoder_layers=16,
                 dim_feedforward=None,
                 dropout=.1,
                 PosEncConstructor=PositionalAppender,
                 ):
        super().__init__()
        self.embedding_dim = embedding_dim
        if dim_feedforward is None:
            dim_feedforward = self.embedding_dim*4
        self.num_agents = num_agents
        self.num_tokens = num_agents + 3
        # adding [MASK], [CLS], and [CLS2] tokens ([CLS] is for target (team) sequences, [CLS2] is for inputs)
        self.CLS = num_agents
        self.CLS2 = num_agents + 1
        self.MASK = num_agents + 2
        self.agent_embedding = Embedding(num_embeddings=self.num_tokens,
                                         embedding_dim=embedding_dim,
                                         )
        self.pos_encoder = PosEncConstructor(d_model=embedding_dim,
                                             dropout=dropout,
                                             )
        self.transform = Transformer(d_model=embedding_dim,
                                     nhead=nhead,
                                     num_encoder_layers=num_encoder_layers,
                                     num_decoder_layers=num_decoder_layers,
                                     dim_feedforward=dim_feedforward,
                                     batch_first=True,  # takes in (N,S,E) input where N is batch and S is seq length
                                     dropout=dropout,
                                     )
        self.output_layer = nn.Linear(embedding_dim, num_agents)
        self.softmax = nn.Softmax(dim=-1)

    def add_cls_tokens(self, target_team):
        """
        adds [CLS] tokens to target teams
        Args:
            target_team: (N, T) vector
        Returns:
            (N, T+1) vector where self.CLS is added to the end of every team
        """
        (N, T) = target_team.shape
        return torch.cat((target_team, torch.ones((N, 1), dtype=target_team.dtype)*self.CLS), dim=1)

    def forward(self, input_embedding, target_team, input_mask, output_probs=True, pre_softmax=False):
        """
        Args:
            input_embedding: (N, S, E) shape tensor of input, or None if no input
                S should be very small, probably the output of embeddings with a more efficient method like LSTM
            target_team: (N, T) shape tensor of team members
                EACH TEAM SHOULD END WITH A [CLS] token
            input_mask: (N, S) tensor of booleans on whether to mask each input
            output_probs: whether to output the probability of each team member
                otherwise just outputs the final embedding
            pre_softmax: if True, does not apply softmax to logits
        Returns:
            if output_probs, (N, T, num_agents) probability distribution for each position
            otherwise, (N, T, embedding_dim) output of transformer model
        """
        N, T = target_team.shape

        # creates a sequence of size S+1
        # dimensions (N, S+1, E) where we add embeddings of [CLS2] tokens for the last values
        if input_embedding is None:
            source = self.agent_embedding(torch.fill(torch.zeros((N, 1), dtype=target_team.dtype), self.CLS2))
            S = 0
        else:
            source = torch.cat((
                input_embedding,
                self.agent_embedding(torch.fill(torch.zeros((N, 1), dtype=target_team.dtype), self.CLS2)),), dim=1)
            N, S, _ = input_embedding.shape
        if input_mask is None:
            input_mask = torch.zeros((N, S), dtype=torch.bool)
        input_mask = torch.cat((input_mask, torch.zeros((N, 1), dtype=torch.bool)), dim=1)

        target = self.agent_embedding(target_team)
        pos_enc_target = self.pos_encoder(target)

        output = self.transform.forward(src=source,
                                        tgt=pos_enc_target,
                                        src_key_padding_mask=input_mask,
                                        memory_key_padding_mask=input_mask,
                                        )
        if output_probs:
            output = self.output_layer(output)
            if not pre_softmax:
                output = self.softmax(output)

        return output


class TeamBuilder(nn.Module):
    def __init__(self,
                 input_embedder: InputEmbedder,
                 berteam: BERTeam,
                 ):
        """
        one of berteam or num_agents must be defined
        Args:
            input_embedder:
            berteam:
            num_agents:
        """
        super().__init__()
        self.input_embedder = input_embedder
        self.berteam = berteam

    def forward(self, obs_preembed, target_team, obs_mask, output_probs=True, pre_softmax=False):
        """
        Args:
            obs_preembed: (N, S, *) shape tensor of input, or None if no input
            target_team: (N, T) shape tensor of team members
                EACH TEAM SHOULD END WITH A [CLS] token
            obs_mask: (N, S) tensor of booleans on whether to mask each input
            output_probs: whether to output the probability of each team member
                otherwise just outputs the final embedding
            pre_softmax: if True, does not apply softmax to logits
        Returns:
            if output_probs, (N, T, num_agents) probability distribution for each position
            otherwise, (N, T, embedding_dim) output of transformer model
        """
        if obs_preembed is None:
            obs_embed = None
            embed_mask = None
        else:
            obs_embed, embed_mask = self.input_embedder(obs_preembed, obs_mask)
        return self.berteam.forward(input_embedding=obs_embed,
                                    target_team=target_team,
                                    input_mask=embed_mask,
                                    output_probs=output_probs,
                                    pre_softmax=pre_softmax,
                                    )


if __name__ == '__main__':
    N = 10
    S = 7
    T = 5
    E = 16

    test = Transformer(d_model=E, batch_first=True)
    s = torch.rand((N, S, E))
    s[0, 0, :] = 100.

    t = torch.rand((N, T, E))
    test.forward(src=s, tgt=t)

    test = BERTeam(num_agents=69, embedding_dim=E, dropout=0)

    team = torch.randint(0, test.num_agents, (N, T))
    team = test.add_cls_tokens(team)

    obs_preembed = torch.rand((N, S, E))
    obs_mask = torch.ones((N, S), dtype=torch.bool)
    obs_mask[:, 0] = False

    a = test.forward(input_embedding=obs_preembed,
                     target_team=team,
                     input_mask=obs_mask
                     )
    obs_preembed[:, 1] *= 10
    b = test.forward(input_embedding=obs_preembed,
                     target_team=team,
                     input_mask=obs_mask,
                     )

    assert torch.all(a == b)



from BERTeam.networks.positional_encoder import IdentityEncoding, ClassicPositionalEncoding, PositionalAppender
from BERTeam.networks.input_embedding import (InputEmbedder,
                                              DiscreteInputEmbedder,
                                              DiscreteInputPosEmbedder,
                                              DiscreteInputPosAppender,
                                              LSTEmbedding,
                                              )
from BERTeam.networks.team_builder import TeamBuilder, BERTeam

__all__ = ['InputEmbedder',
           'DiscreteInputEmbedder',
           'DiscreteInputPosAppender',
           'DiscreteInputPosEmbedder',
           'LSTEmbedding',

           'IdentityEncoding',
           'ClassicPositionalEncoding',
           'PositionalAppender',
           'TeamBuilder',
           'BERTeam'
           ]



from BERTeam.buffer.language_replay_buffer import (LangReplayBuffer,
                                                   ReplayBufferDiskStorage,
                                                   BinnedReplayBufferDiskStorage,
                                                   )

__all__ = ['LangReplayBuffer',
           'ReplayBufferDiskStorage',
           'BinnedReplayBufferDiskStorage',
           ]



import torch
from BERTeam.trainer import DiscreteInputTrainer
from BERTeam.buffer import ReplayBufferDiskStorage
from matplotlib import pyplot as plt
from BERTeam.tests.rps_basic.game import SingleOutcome, plot_dist_evolution


def loss_plot(losses, save_dir=None, show=False):
    plt.plot(losses)
    plt.xlabel('Epoch')
    plt.ylabel('Loss')

    plt.title('CrossEntropy Loss')
    if save_dir is not None:
        plt.savefig(save_dir, bbox_inches='tight')
    if show:
        plt.show()
    plt.close()


if __name__ == '__main__':
    import os, sys

    torch.random.manual_seed(69)

    DIR = os.getcwd()
    plot_dir = os.path.join(DIR, 'data', 'plots', 'tests_rps_teams')
    if not os.path.exists((plot_dir)):
        os.makedirs(plot_dir)
    trainer = DiscreteInputTrainer(
        buffer=ReplayBufferDiskStorage(storage_dir=os.path.join(DIR, 'data', 'test_storage'),
                                       capacity=1e5,
                                       device=None,
                                       ),
        num_agents=3,
        num_input_tokens=3,
        embedding_dim=64,
        pos_encode_input=False,
        num_decoder_layers=4,
        num_encoder_layers=4,
    )

    minibatch = 64
    init_dists = []
    cond_dists = []
    losses = []
    outcome = SingleOutcome(torch.arange(3))
    for epoch in range(100):
        noise = trainer.create_nose_model_towards_uniform(.1)
        # create two teams of 1 agent, where an agent is an integer from 0 to 2
        for _ in range(minibatch):
            game = torch.randint(0, 3, (2, 1))
            trainer.buffer.add_outcome(teams=game,
                                       outcome=outcome.get_outcome(team_choices=game, agent_choices=None),
                                       filter=lambda score: score > .5,
                                       )

            # trainer.buffer.add_outcome(teams=game,
            #                           outcome=outcome.get_outcome(team_choices=game, agent_choices=None),
            #                           filter=lambda score: score > .5,
            #                           ignore_observation=True
            #                           )

        init_distribution = trainer.team_builder.forward(obs_preembed=None,
                                                         target_team=trainer.create_masked_teams(T=1, N=1),
                                                         obs_mask=None,
                                                         ).detach().flatten().numpy()
        init_dists.append(init_distribution)
        conditional_dists = []
        for opponent in range(3):
            dist = trainer.team_builder.forward(obs_preembed=torch.tensor([[opponent]]),
                                                target_team=trainer.create_masked_teams(T=1, N=1),
                                                obs_mask=None,
                                                )
            conditional_dists.append(dist.detach().flatten().numpy())
        cond_dists.append(conditional_dists)
        loss = trainer.train(batch_size=minibatch)
        losses.append(loss)
        print('epoch', epoch, ';\tbuffer size', len(trainer.buffer))
        print('loss', loss)
        if not (epoch + 1)%10:
            loss_plot(losses=losses, save_dir=os.path.join(plot_dir, 'loss_plot.png'))
            plot_dist_evolution(init_dists, save_dir=os.path.join(plot_dir, 'init_dist.png'))
            for k, name in enumerate('RPS'):
                plot_dist_evolution([dist[k] for dist in cond_dists],
                                    save_dir=os.path.join(plot_dir, 'dist_against' + name + '.png'),
                                    title='Distribution against ' + name)
    trainer.clear()



import torch
from matplotlib import pyplot as plt
from BERTeam.outcome import PlayerInfo, OutcomeFn

ROCK = 0
PAPER = 1
SCISOR = 2


def double_game_outcome(a, b):
    # ['RR', 'PP', 'SS', 'RP', 'RS', 'PS']
    a, b = int(a), int(b)

    if a < 3 and b < 3:
        if a == b:
            return 0
        elif (a - b)%3 == 1:
            # i wins
            return 1
        else:
            return -1
    if a >= 3 and b >= 3:
        # ['RR', 'PP', 'SS', 'RP', 'RS', 'PS']
        # mapped to
        # ['PS', 'RS', 'RP', ...]
        # equivalent to
        # [ R P S ]
        return double_game_outcome(5 - a, 5 - b)
    elif ({a, b} in ({0, 4}, {1, 3}, {2, 5})):
        if a < b:
            return 1
        else:
            return -1
    else:
        if a > b:
            return 1
        else:
            return -1


class DualPreMappedOutcome(OutcomeFn):
    def __init__(self, agents):
        super().__init__()
        self.agents = agents

    def get_outcome(self, team_choices, **kwargs):
        result = double_game_outcome(team_choices[0], team_choices[1])
        if result == 0:
            return [(.5, [PlayerInfo(obs_preembed=team_choices[1]),
                          PlayerInfo(obs_preembed=team_choices[1])]),
                    (.5, [PlayerInfo(obs_preembed=team_choices[0]),
                          PlayerInfo(obs_preembed=team_choices[0])]),
                    ]
        if result == 1:
            return [(1, [PlayerInfo(obs_preembed=team_choices[1]),
                         PlayerInfo(obs_preembed=team_choices[1])]),
                    (0, [PlayerInfo(obs_preembed=team_choices[0]),
                         PlayerInfo(obs_preembed=team_choices[0])]),
                    ]
        if result == -1:
            return [(0, [PlayerInfo(obs_preembed=team_choices[1]),
                         PlayerInfo(obs_preembed=team_choices[1])]),
                    (1, [PlayerInfo(obs_preembed=team_choices[0]),
                         PlayerInfo(obs_preembed=team_choices[0])]),
                    ]


class DualPairOutcome(OutcomeFn):

    def map_to_number(self, team):
        # ['RR', 'PP', 'SS', 'RP', 'RS', 'PS']
        team = team.flatten()
        if team[0] == team[1]:
            return team[0]
        team = sorted(team)
        if team == [0, 1]:
            return torch.tensor(3)
        if team == [1, 2]:
            return torch.tensor(5)
        return torch.tensor(4)

    def get_outcome(self, team_choices, **kwargs):
        result = double_game_outcome(self.map_to_number(team_choices[0]),
                                     self.map_to_number(team_choices[1]),
                                     )
        if result == 0:
            return [(.5, [PlayerInfo(obs_preembed=team_choices[1]),
                          PlayerInfo(obs_preembed=team_choices[1])]),
                    (.5, [PlayerInfo(obs_preembed=team_choices[0]),
                          PlayerInfo(obs_preembed=team_choices[0])]),
                    ]
        if result == 1:
            return [(1, [PlayerInfo(obs_preembed=team_choices[1]),
                         PlayerInfo(obs_preembed=team_choices[1])]),
                    (0, [PlayerInfo(obs_preembed=team_choices[0]),
                         PlayerInfo(obs_preembed=team_choices[0])]),
                    ]
        if result == -1:
            return [(0, [PlayerInfo(obs_preembed=team_choices[1]),
                         PlayerInfo(obs_preembed=team_choices[1])]),
                    (1, [PlayerInfo(obs_preembed=team_choices[0]),
                         PlayerInfo(obs_preembed=team_choices[0])]),
                    ]


if __name__ == '__main__':
    torch.random.manual_seed(69)
    agents = torch.arange(6)
    outcomes = DualPreMappedOutcome(agents=agents)
    print(outcomes.get_outcome(team_choices=(torch.tensor([3]), torch.tensor([0]))))
    print(outcomes.get_outcome(team_choices=(torch.tensor([3]), torch.tensor([1]))))
    print(outcomes.get_outcome(team_choices=(torch.tensor([3]), torch.tensor([2]))))
    print()
    outcomes = DualPairOutcome(agents=agents)
    print(outcomes.get_outcome(team_choices=(torch.tensor([2, 1]), torch.tensor([1, 1]))))



import torch
from matplotlib import pyplot as plt
from BERTeam.outcome import PlayerInfo, OutcomeFn

ROCK = 0
PAPER = 1
SCISOR = 2


def plot_dist_evolution(plot_dist, save_dir=None, show=False, alpha=1, labels='RPS', title=None):
    num_pops = len(plot_dist[0])
    x = range(len(plot_dist))
    for i in range(num_pops):
        plt.fill_between(x=x,
                         y1=[sum(dist[:i]) for dist in plot_dist],
                         y2=[sum(dist[:i + 1]) for dist in plot_dist],
                         label=labels[i] if labels is not None and len(labels) >= i else None,
                         alpha=alpha,
                         )
    if title is not None:
        plt.title(title)
    plt.legend(loc='center left', bbox_to_anchor=(-.25, .5))

    if save_dir is not None:
        plt.savefig(save_dir, bbox_inches='tight')
    if show:
        plt.show()
    plt.close()


class SingleOutcome(OutcomeFn):
    def __init__(self, agents):
        super().__init__()
        self.agents = agents

    def get_outcome(self, team_choices, agent_choices, **kwargs):
        i, j = team_choices
        diff = (self.agents[i[0]] - self.agents[j[0]])%3
        if diff == 0:  # the agents tied
            choice = self.agents[i,]
            return [(.5, [PlayerInfo(obs_preembed=choice)]),
                    (.5, [PlayerInfo(obs_preembed=choice)])]
        if diff == 1:
            # agent i won
            return [
                (1, [PlayerInfo(obs_preembed=self.agents[j,])]),
                (0, [PlayerInfo(obs_preembed=self.agents[i,])]),
            ]

        if diff == 2:
            # agent j won
            return [
                (0, [PlayerInfo(obs_preembed=self.agents[j,])]),
                (1, [PlayerInfo(obs_preembed=self.agents[i,])]),
            ]


if __name__ == '__main__':
    torch.random.manual_seed(69)
    agents = torch.arange(3)
    outcomes = SingleOutcome(agents=agents)
    print(outcomes.get_outcome(team_choices=(torch.tensor([0]), torch.tensor([1])), agent_choices=None))
    print(outcomes.get_outcome(team_choices=(torch.tensor([1]), torch.tensor([1])), agent_choices=None))
    print(outcomes.get_outcome(team_choices=(torch.tensor([2]), torch.tensor([1])), agent_choices=None))



import torch


class PlayerInfo:
    """
    structure for player observations of a multi-agent game
    """

    def __init__(self, obs_preembed=None, obs_mask=None):
        """
        Args:
            obs_preembed: (S,*) array of a sequence of observations
                None if no observations
            obs_mask: (S,) boolean array of which observations to mask
                None if no mask
        """
        self.obs_preembed = obs_preembed
        self.obs_mask = obs_mask
        self.S = 0 if obs_preembed is None else obs_preembed.shape[0]

    def clone(self):
        return PlayerInfo(obs_preembed=None if self.obs_preembed is None else self.obs_preembed.clone(),
                          obs_mask=None if self.obs_mask is None else self.obs_mask.clone(),
                          )

    def get_data(self, reshape=True):
        """
        gets data
        Args:
            reshape: whtether to return in batch format
        Returns:
            self.obs_preembed,self.obs_mask
        """
        if reshape:
            return (None if self.obs_preembed is None else self.obs_preembed.unsqueeze(0),
                    None if self.obs_mask is None else self.obs_mask.unsqueeze(0))
        else:
            return self.obs_preembed, self.obs_mask

    def union_obs(self,
                  other_player_info,
                  combine=True,
                  ):
        """
        creates clone, combining observations with other_obs and other_mask

        Args:
            other_player_info: other player info with obs and mask either None or shape (S',*) and (S',)
                if self.obs_preembed has shape (S,*), the two * dimensions must be the same
            combine: whether to combine or just return updates
        Returns:
            concatenates the two preembeddings and masks
            either outputs an outcome with None, None or
                (S+1+S', *) concatenated output, (S+1+S',) boolean mask
                 an extra masked element is added in the middle to denote the division
                or other_obs_preembed and other_mask (if not chain_observations)
        """
        other_obs_preembed = other_player_info.obs_preembed
        other_mask = other_player_info.obs_mask
        if not combine:
            return PlayerInfo(obs_preembed=other_obs_preembed,
                              obs_mask=other_mask,
                              )
        Spp = self.S + 1 + other_player_info.S

        if Spp == 1:
            # in this case, there are no observations and no masks, so just return empty observation
            return PlayerInfo()

        if (self.obs_preembed is not None) and (other_obs_preembed is not None):
            # if both are not None, we append the observations,  size (S+1+S', *)

            # divider is shape (1,*)
            divider = torch.zeros_like(self.obs_preembed[:1])
            new_preembed = torch.cat((self.obs_preembed, divider, other_obs_preembed), dim=0)

            # we must also set the mask
            # by default mask nothing except the divider
            new_mask = torch.zeros(Spp, dtype=torch.bool)
            new_mask[self.S] = True  # always mask the division

            if self.obs_mask is not None:
                new_mask[:self.S] = self.obs_mask
            if other_mask is not None:
                new_mask[self.S + 1:] = other_mask
        else:
            # in this case, one is empty, the other has observations
            # just return the one that is nonempty
            if self.obs_preembed is not None:
                new_preembed = self.obs_preembed.clone()
                new_mask = self.obs_mask
            else:
                new_preembed = other_obs_preembed
                new_mask = other_mask
        return PlayerInfo(
            obs_preembed=new_preembed,
            obs_mask=new_mask,
        )

    def __str__(self):
        return ('PlayerInfo(' +
                'obs_preembed:' + str(self.obs_preembed) + '; ' +
                'obs_mask:' + str(self.obs_mask) + '; ' +
                ')')

    def __eq__(self, other):
        if self.obs_preembed is None or other.obs_preembed is None:
            return (self.obs_preembed is None) and (other.obs_preembed is None)

        mask_equality = (
                (self.obs_mask is None and other.obs_mask is None) or
                (self.obs_mask is None and torch.sum(other.obs_mask) == 0) or
                (other.obs_mask is None and torch.sum(self.obs_mask) == 0) or
                torch.equal(self.obs_mask, other.obs_mask)
        )
        return torch.equal(self.obs_preembed, other.obs_preembed) and mask_equality


class OutcomeFn:
    """
    structure for calculating outcomes of a team game
    gives an outcome score for each team, all outcome scores are non-negative and sum to 1
        usually 1 is win, 0.5 is tie (for two team games) and 0 is loss
    """

    def __init__(self):
        super().__init__()
        self.ident = 0
        self.dir = None

    def get_outcome(self, team_choices, agent_choices, **kwargs):
        """
        Args:
            team_choices: K-tuple of teams, each team is an array of players
            agent_choices: same shape as team_choices, calculated agents (if applicable)

        Returns: list corresponding to teams
            [
                outcome score,
                list corresponding to players of PlayerInfo(
                    obs_preembed=player observation (None or size (S,*) seq of observations);
                    obs_mask=observation mask (None or size (S,) boolean array of which items to mask;
                    )
                list can be empty, this will correspond to an empty observation
            ]
        """
        raise NotImplementedError

    def set_ident(self, ident):
        self.ident = ident

    def set_dir(self, dir):
        self.dir = dir


if __name__ == '__main__':
    test0 = PlayerInfo()
    test = PlayerInfo(obs_preembed=torch.rand((1, 2, 3)))
    test2 = PlayerInfo(obs_preembed=torch.rand((2, 2, 3)))
    print(test0)
    print(test)
    print(test0.union_obs(test))
    print(test.union_obs(test))
    print(test.union_obs(test2))
    print(test0 == test)
    print(test0.union_obs(test) == test)
    print(test.union_obs(test0) == test)



from BERTeam.trainer.team_trainer import (TeamTrainer,
                                          MLMTeamTrainer,
                                          DiscreteInputTrainer,
                                          )

__all__ = ['TeamTrainer',
           'MLMTeamTrainer',
           'DiscreteInputTrainer',
           ]



import torch
from BERTeam.trainer import DiscreteInputTrainer
from BERTeam.buffer import ReplayBufferDiskStorage

import numpy as np
import itertools

ROCK = 0
PAPER = 1
SCISOR = 2


def dist_from_trainer(trainer: DiscreteInputTrainer,
                      input_preembedding=None,
                      input_mask=None,
                      num_agents=3,
                      keyorder=None
                      ):
    trainer.team_builder.eval()
    dic = dict()
    keys = []
    for choice in itertools.chain(itertools.combinations(range(num_agents), 1),
                                  itertools.combinations(range(num_agents), 2)):
        if len(choice) == 1:
            choice = choice + choice
            perms = [choice]
        else:
            perms = itertools.permutations(choice)
        keys.append(choice)
        prob = 0
        for perm in perms:
            orders = list(itertools.permutations(range(2)))
            for order in orders:
                this_prob = 1.
                target_team = trainer.create_masked_teams(T=2, N=1)

                for idx in order:
                    full_dist = trainer.team_builder.forward(input_preembedding, target_team, input_mask)
                    dist = full_dist[0, idx]
                    this_prob *= dist[perm[idx]].item()
                    target_team[0, idx] = perm[idx]
                prob += this_prob/len(orders)
        dic[choice] = prob

    trainer.team_builder.train()
    if keyorder is None:
        keyorder = keys
    return np.array([dic[key] for key in keyorder])


if __name__ == '__main__':
    import os, sys
    from BERTeam.tests.rps_basic.teams import plot_dist_evolution, loss_plot
    from BERTeam.tests.rps_dual.game import DualPairOutcome
    import time

    torch.random.manual_seed(69)
    agents = torch.arange(3)

    DIR = os.getcwd()
    plot_dir = os.path.join(DIR, 'data', 'plots', 'tests_rps2_teams')
    if not os.path.exists((plot_dir)):
        os.makedirs(plot_dir)
    trainer = DiscreteInputTrainer(num_agents=3,
                                   num_input_tokens=3,
                                   embedding_dim=64,
                                   pos_encode_input=True,
                                   append_pos_encode_input=True,
                                   pos_encode_teams=True,
                                   append_pos_encode_teams=True,
                                   num_decoder_layers=4,
                                   num_encoder_layers=4,
                                   buffer=ReplayBufferDiskStorage(
                                       storage_dir=os.path.join(DIR, "data", "temp", "tests_rps2_teams"),
                                       capacity=int(2e4),
                                       device=None,
                                   ),
                                   )

    N = 100
    minibatch = 64
    init_dists = []
    cond_dists = []
    strat_labels = ['RR', 'PP', 'SS', 'RP', 'RS', 'PS']
    losses = []
    outcome = DualPairOutcome()
    for epoch in range(100):
        start_time = time.time()
        noise = trainer.create_nose_model_towards_uniform(1/np.sqrt(epoch/2 + 1))
        game = trainer.create_teams(T=2, N=2, noise_model=noise)


        trainer.buffer.add_outcome(teams=game,
                                   outcome=outcome.get_outcome(team_choices=game),
                                   filter=lambda score: score > .5,
                                   )

        init_distribution = dist_from_trainer(trainer=trainer,
                                              input_preembedding=None,
                                              input_mask=None,
                                              )
        init_dists.append(init_distribution)
        conditional_dists = []
        for opponent_choice in itertools.chain(itertools.combinations(range(3), 1),
                                               itertools.combinations(range(3), 2)):
            if len(opponent_choice) == 1:
                opponent_choice = opponent_choice + opponent_choice
            dist = dist_from_trainer(trainer=trainer,
                                     input_preembedding=torch.tensor([opponent_choice]),
                                     input_mask=None,
                                     )
            conditional_dists.append(dist)
        cond_dists.append(conditional_dists)

        collection_time = time.time() - start_time
        start_time = time.time()
        loss = trainer.train(batch_size=minibatch)
        losses.append(loss)

        print('epoch', epoch, ';\tbuffer size', len(trainer.buffer))
        print('\tcollection time:', round(collection_time, 2))
        print('\ttrain time:', round(time.time() - start_time, 2))

        print('\tloss', loss)

        if not (epoch + 1)%10:
            start_time = time.time()
            loss_plot(losses, save_dir=os.path.join(plot_dir, 'loss_plot.png'))

            plot_dist_evolution(init_dists,
                                save_dir=os.path.join(plot_dir, 'init_dist.png'),
                                labels=strat_labels
                                )
            for k, name in enumerate(strat_labels):
                plot_dist_evolution([dist[k] for dist in cond_dists],
                                    save_dir=os.path.join(plot_dir, 'dist_against' + name + '.png'),
                                    title='Distribution against ' + name,
                                    labels=strat_labels,
                                    )
            print('\tplot time:', round(time.time() - start_time, 2))
        epoch += 1
        print()
    trainer.clear()
